#!/usr/bin/env python3

from bunkai.algorithm.bunkai_sbd.bunkai_sbd import BunkaiSentenceBoundaryDisambiguation as Bunkai

__all__ = ["Bunkai"]
__version_info__ = (1, 5, 7)
__version__ = ".".join(map(str, __version_info__))



#!/usr/bin/env python3
from bunkai.algorithm.tsunoda_sbd.annotator.basic_annotator import BasicRule
from bunkai.algorithm.tsunoda_sbd.annotator.exception_no import ExceptionNo
from bunkai.algorithm.tsunoda_sbd.annotator.exception_numeric import ExceptionNumeric
from bunkai.algorithm.tsunoda_sbd.annotator.exception_particle import ExceptionParticle
from bunkai.algorithm.tsunoda_sbd.annotator.morph_annotator_janome import MorphAnnotatorJanome
from bunkai.algorithm.tsunoda_sbd.annotator.replace_parentheses import ExceptionParentheses

__all__ = [
    "BasicRule",
    "ExceptionNo",
    "ExceptionNumeric",
    "ExceptionParticle",
    "MorphAnnotatorJanome",
    "ExceptionParentheses",
]



#!/usr/bin/env python3

import collections
import logging
import os
import typing
import unicodedata

from janome.tokenizer import Tokenizer
from transformers.models.bert.tokenization_bert import BertTokenizer, WordpieceTokenizer, load_vocab
from transformers.utils.hub import cached_file

import bunkai.constant

"""
The original source code is from cl-tohoku/bert-japanese.
https://github.com/cl-tohoku/bert-japanese/blob/master/tokenization.py
The original source code is under Apache-2.0 License.
"""

logger = logging.getLogger(__name__)


class JanomeTokenizer(object):
    """Runs basic tokenization with Janome morphological parser."""

    def __init__(self, *, do_lower_case=False, never_split=None, normalize_text=True):
        """
        Construct a JanomeTokenizer.

        :arg do_lower_case: (`optional`) boolean (default True)
                Whether to lower case the input.
        :arg never_split: (`optional`) list of str
                Kept for backward compatibility purposes.
                Now implemented directly at the base class level (see :func:`PreTrainedTokenizer.tokenize`)
                List of token not to split.
        :arg normalize_text: (`optional`) boolean (default True)
                Whether to apply unicode normalization to text before tokenization.
        """
        self.do_lower_case = do_lower_case
        self.never_split = never_split if never_split is not None else []
        self.normalize_text = normalize_text
        self.janome_tokenizer = Tokenizer()

    def tokenize(self, text: str, *, never_split=None, **kwargs):
        """Tokenizes a piece of text."""
        if self.normalize_text:
            text = unicodedata.normalize("NFKC", text)

        never_split = self.never_split + (never_split if never_split is not None else [])
        tokens = self.janome_tokenizer.tokenize(text)
        __tokens = []
        last_index = 0
        for t in tokens:
            token = t.surface
            token_start = text.index(token, last_index)
            if last_index != token_start:
                __tokens.append(text[last_index:token_start])

            if self.do_lower_case and token not in never_split:
                token = token.lower()
                __tokens.append(token.lower())
            else:
                __tokens.append(token)
            last_index = token_start + len(token)

        if len(text) != last_index:
            __tokens.append(text[last_index:])

        assert text == "".join(__tokens), f"[{text}] != [{''.join(__tokens)}]"
        return __tokens


class CharacterTokenizer(object):
    """Runs Character tokenziation."""

    def __init__(self, vocab, unk_token, normalize_text=True):
        self.vocab = vocab
        self.unk_token = unk_token
        self.normalize_text = normalize_text

    def tokenize(self, text):
        """
        Tokenize a piece of text into characters.

        For example:
            input = "apple"
            output = ["a", "p", "p", "l", "e"]
        :arg text: A single token or whitespace separated tokens.
        This should have already been passed through `BasicTokenizer`.
        :return: A list of characters.
        """
        if self.normalize_text:
            text = unicodedata.normalize("NFKC", text)

        output_tokens = []
        for char in text:
            if char not in self.vocab:
                output_tokens.append(self.unk_token)
                continue

            output_tokens.append(char)

        return output_tokens


class JanomeSubwordsTokenizer(BertTokenizer):
    def __init__(
        self,
        vocab_file,
        *,
        subword_tokenizer_type="wordpiece",
        do_subword_tokenize: bool = True,
        never_split=None,
        unk_token="[UNK]",
        sep_token="[SEP]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        mask_token="[MASK]",
        **kwargs,
    ):
        """
        Construct a JanomeSubwordsTokenizer.

        :arg vocab_file: Path to a one-wordpiece-per-line vocabulary file.
        :arg do_lower_case: (`optional`) boolean (default True)
        Whether to lower case the input.
        Only has an effect when do_basic_tokenize=True.
        :arg do_word_tokenize: (`optional`) boolean (default True) Whether to do word tokenization.
        :arg do_subword_tokenize: (`optional`) boolean (default True) Whether to do subword tokenization.
        :arg word_tokenizer_type: (`optional`) string (default "basic")
                Type of word tokenizer. basic / janome / pre_tokenize
        :arg subword_tokenizer_type: (`optional`) string (default "wordpiece") Type of subword tokenizer.
        :arg cls_token: No description.
        """
        super(BertTokenizer, self).__init__(
            unk_token=unk_token,
            sep_token=sep_token,
            pad_token=pad_token,
            cls_token=cls_token,
            mask_token=mask_token,
            **kwargs,
        )

        if os.path.isfile(vocab_file):
            self.vocab = load_vocab(vocab_file)
        else:
            self.vocab = load_vocab(cached_file(vocab_file, "vocab.txt"))

        # add new vocab
        self.add_tokens([" ", bunkai.constant.METACHAR_LINE_BREAK])

        self.ids_to_tokens = collections.OrderedDict([(ids, tok) for tok, ids in self.vocab.items()])

        self.do_word_tokenize = False
        self.do_subword_tokenize = True
        if do_subword_tokenize:
            if subword_tokenizer_type == "wordpiece":
                self.subword_tokenizer = WordpieceTokenizer(vocab=self.vocab, unk_token=self.unk_token)
            elif subword_tokenizer_type == "character":
                self.subword_tokenizer = CharacterTokenizer(vocab=self.vocab, unk_token=self.unk_token)
            else:
                raise ValueError("Invalid subword_tokenizer_type '{}' is specified.".format(subword_tokenizer_type))

        self.janome_tokenizer = JanomeTokenizer()

    def tokenize(self, text: typing.Union[str, typing.List[str]]) -> typing.List[str]:
        if isinstance(text, str):
            morphemes = self.janome_tokenizer.tokenize(text)
        elif isinstance(text, list) and all([isinstance(t, str) for t in text]):
            morphemes = text
        else:
            raise Exception(f"Invalid input-type {text}")

        if self.do_subword_tokenize:
            split_tokens = []
            for token in morphemes:
                sts = [sub_token for sub_token in self.subword_tokenizer.tokenize(token)]
                if len(sts) == 0:
                    split_tokens.append(token)
                else:
                    split_tokens += sts
        else:
            split_tokens = morphemes

        return split_tokens



# mypy: ignore-errors
# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Named entity recognition fine-tuning: utilities to work with CoNLL-2003 task.


import logging
import os
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Union

from filelock import FileLock
from transformers import PreTrainedTokenizer, is_tf_available, is_torch_available

from bunkai.algorithm.lbd.custom_tokenizers import JanomeSubwordsTokenizer

logger = logging.getLogger(__name__)


@dataclass
class InputExample:
    """
    A single training/test example for token classification.

    :param guid: Unique id for the example.
    :param words: list. The words of the sequence.
    :param labels: (Optional) list. The labels for each word of the sequence. This should be
    :param specified for train and dev examples, but not for test examples.
    :param is_document_first: First fragment of document.
    """

    guid: str
    words: List[str]
    labels: Optional[List[str]]
    is_document_first: bool


@dataclass
class InputFeatures:
    """
    A single set of features of data.

    Property names are the same names as the corresponding inputs to a model.
    """

    input_ids: List[int]
    attention_mask: List[int]
    token_type_ids: Optional[List[int]] = None
    label_ids: Optional[List[int]] = None
    document_id: Optional[str] = None


class Split(Enum):
    train = "train"
    dev = "dev"
    test = "test"


if is_torch_available():
    import torch
    from torch import nn
    from torch.utils.data.dataset import Dataset

    class NerDataset(Dataset):
        features: List[InputFeatures]
        pad_token_label_id: int = nn.CrossEntropyLoss().ignore_index
        # Use cross entropy ignore_index as padding label id so that only
        # real label ids contribute to the loss later.

        def __init__(
            self,
            data_dir: str,
            tokenizer: JanomeSubwordsTokenizer,
            labels: List[str],
            model_type: str,
            max_seq_length: Optional[int] = None,
            overwrite_cache=False,
            mode: Split = Split.train,
            is_distil_bert: bool = False,
        ):
            # Load data features from cache or dataset file
            cached_features_file = os.path.join(
                data_dir,
                "cached_{}_{}_{}".format(mode.value, tokenizer.__class__.__name__, str(max_seq_length)),
            )

            # Make sure only the first process in distributed training processes the dataset,
            # and the others will use the cache.
            lock_path = cached_features_file + ".lock"
            with FileLock(lock_path):
                if os.path.exists(cached_features_file) and not overwrite_cache:
                    logger.info(f"Loading features from cached file {cached_features_file}")
                    self.features = torch.load(cached_features_file)
                else:
                    logger.info(f"Creating features from dataset file at {data_dir}")
                    examples = read_examples_from_file(data_dir, mode)
                    # TODO clean up all this to leverage built-in features of tokenizers
                    self.features = convert_examples_to_features(
                        examples=examples,
                        label_list=labels,
                        max_seq_length=max_seq_length,
                        tokenizer=tokenizer,
                        cls_token_at_end=bool(model_type in ["xlnet"]),
                        # xlnet has a cls token at the end
                        cls_token=tokenizer.cls_token,
                        cls_token_segment_id=2 if model_type in ["xlnet"] else 0,
                        sep_token=tokenizer.sep_token,
                        sep_token_extra=False,
                        # roberta uses an extra separator b/w pairs of sentences,
                        # cf. github.com/pytorch/fairseq/commit/1684e166e3da03f5b600dbb7855cb98ddfcd0805
                        pad_on_left=bool(tokenizer.padding_side == "left"),
                        pad_token=tokenizer.pad_token_id,
                        pad_token_segment_id=tokenizer.pad_token_type_id,
                        pad_token_label_id=self.pad_token_label_id,
                        is_distil_bert=is_distil_bert,
                    )
                    logger.info(f"Saving features into cached file {cached_features_file}")
                    torch.save(self.features, cached_features_file)

        def __len__(self):
            return len(self.features)

        def __getitem__(self, i) -> InputFeatures:
            return self.features[i]


if is_tf_available():
    import tensorflow as tf

    class TFNerDataset:
        features: List[InputFeatures]
        pad_token_label_id: int = -100
        # Use cross entropy ignore_index as padding label id so that only
        # real label ids contribute to the loss later.

        def __init__(
            self,
            data_dir: str,
            tokenizer: PreTrainedTokenizer,
            labels: List[str],
            model_type: str,
            max_seq_length: Optional[int] = None,
            overwrite_cache=False,
            mode: Split = Split.train,
        ):
            examples = read_examples_from_file(data_dir, mode)
            # TODO clean up all this to leverage built-in features of tokenizers
            self.features = convert_examples_to_features(
                examples,
                labels,
                max_seq_length,
                tokenizer,
                cls_token_at_end=bool(model_type in ["xlnet"]),
                # xlnet has a cls token at the end
                cls_token=tokenizer.cls_token,
                cls_token_segment_id=2 if model_type in ["xlnet"] else 0,
                sep_token=tokenizer.sep_token,
                sep_token_extra=False,
                # roberta uses an extra separator b/w pairs of sentences,
                # cf. github.com/pytorch/fairseq/commit/1684e166e3da03f5b600dbb7855cb98ddfcd0805
                pad_on_left=bool(tokenizer.padding_side == "left"),
                pad_token=tokenizer.pad_token_id,
                pad_token_segment_id=tokenizer.pad_token_type_id,
                pad_token_label_id=self.pad_token_label_id,
            )

            def gen():
                for ex in self.features:
                    if ex.token_type_ids is None:
                        yield (
                            {
                                "input_ids": ex.input_ids,
                                "attention_mask": ex.attention_mask,
                            },
                            ex.label_ids,
                        )
                    else:
                        yield (
                            {
                                "input_ids": ex.input_ids,
                                "attention_mask": ex.attention_mask,
                                "token_type_ids": ex.token_type_ids,
                            },
                            ex.label_ids,
                        )

            if "token_type_ids" not in tokenizer.model_input_names:
                self.dataset = tf.data.Dataset.from_generator(
                    gen,
                    ({"input_ids": tf.int32, "attention_mask": tf.int32}, tf.int64),
                    (
                        {
                            "input_ids": tf.TensorShape([None]),
                            "attention_mask": tf.TensorShape([None]),
                        },
                        tf.TensorShape([None]),
                    ),
                )
            else:
                self.dataset = tf.data.Dataset.from_generator(
                    gen,
                    (
                        {
                            "input_ids": tf.int32,
                            "attention_mask": tf.int32,
                            "token_type_ids": tf.int32,
                        },
                        tf.int64,
                    ),
                    (
                        {
                            "input_ids": tf.TensorShape([None]),
                            "attention_mask": tf.TensorShape([None]),
                            "token_type_ids": tf.TensorShape([None]),
                        },
                        tf.TensorShape([None]),
                    ),
                )

        def get_dataset(self):
            self.dataset = self.dataset.apply(tf.data.experimental.assert_cardinality(len(self.features)))

            return self.dataset

        def __len__(self):
            return len(self.features)

        def __getitem__(self, i) -> InputFeatures:
            return self.features[i]


def read_examples_from_file(data_dir, mode: Union[Split, str]) -> List[InputExample]:
    if isinstance(mode, Split):
        mode = mode.value
    file_path = os.path.join(data_dir, f"{mode}.txt")
    guid_index = 1
    examples = []
    with open(file_path, encoding="utf-8") as f:
        words = []
        labels = []
        for line in f:
            if line.startswith("-DOCSTART-") or line == "" or line == "\n":
                if words:
                    examples.append(
                        InputExample(
                            guid=f"{mode}-{guid_index}",
                            words=words,
                            labels=labels,
                            is_document_first=True,
                        )
                    )
                    guid_index += 1
                    words = []
                    labels = []
            else:
                splits = line.rsplit(" ", 1)
                words.append(splits[0])
                if len(splits) > 1:
                    labels.append(splits[-1].replace("\n", ""))
                else:
                    # Examples could have no label for mode = "test"
                    labels.append("O")
        if words:
            examples.append(
                InputExample(
                    guid=f"{mode}-{guid_index}",
                    words=words,
                    labels=labels,
                    is_document_first=True,
                )
            )
    return examples


def convert_examples_to_features(
    examples: List[InputExample],
    label_list: List[str],
    max_seq_length: int,
    tokenizer: JanomeSubwordsTokenizer,
    cls_token_at_end=False,
    cls_token="[CLS]",
    cls_token_segment_id=1,
    sep_token="[SEP]",
    sep_token_extra=False,
    pad_on_left=False,
    pad_token=0,
    pad_token_segment_id=0,
    pad_token_label_id=-100,
    sequence_a_segment_id=0,
    mask_padding_with_zero=True,
    is_distil_bert: bool = False,
) -> List[InputFeatures]:
    """
    Load a data file into a list of `InputFeatures`.

    :param cls_token_at_end: define the location of the CLS token:
        - False (Default, BERT/XLM pattern): [CLS] + A + [SEP] + B + [SEP]
        - True (XLNet/GPT pattern): A + [SEP] + B + [SEP] + [CLS]
    :param cls_token_segment_id: define the segment id associated to the CLS token (0 for BERT, 2 for XLNet)
    """
    # TODO clean up all this to leverage built-in features of tokenizers

    label_map = {label: i for i, label in enumerate(label_list)}

    features = []
    for ex_index, example in enumerate(examples):
        if ex_index % 10_000 == 0:
            logger.info("Writing example %d of %d", ex_index, len(examples))

        tokens = []
        label_ids = []
        for word, label in zip(example.words, example.labels):
            word_tokens = tokenizer.tokenize(word)
            # bert-base-multilingual-cased sometimes output "nothing ([]) when calling tokenize with just a space.
            if len(word_tokens) > 0:
                tokens.extend(word_tokens)
                # Use the real label id for the first token of the word, and padding ids for the remaining tokens
                label_ids.extend([label_map[label]] + [pad_token_label_id] * (len(word_tokens) - 1))

        # Account for [CLS] and [SEP] with "- 2" and with "- 3" for RoBERTa.
        special_tokens_count = tokenizer.num_special_tokens_to_add()
        if len(tokens) > max_seq_length - special_tokens_count:
            tokens = tokens[: (max_seq_length - special_tokens_count)]
            label_ids = label_ids[: (max_seq_length - special_tokens_count)]

        # The convention in BERT is:
        # (a) For sequence pairs:
        #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
        #  type_ids:   0   0  0    0    0     0       0   0   1  1  1  1   1   1
        # (b) For single sequences:
        #  tokens:   [CLS] the dog is hairy . [SEP]
        #  type_ids:   0   0   0   0  0     0   0
        #
        # Where "type_ids" are used to indicate whether this is the first
        # sequence or the second sequence. The embedding vectors for `type=0` and
        # `type=1` were learned during pre-training and are added to the wordpiece
        # embedding vector (and position vector). This is not *strictly* necessary
        # since the [SEP] token unambiguously separates the sequences, but it makes
        # it easier for the model to learn the concept of sequences.
        #
        # For classification tasks, the first vector (corresponding to [CLS]) is
        # used as as the "sentence vector". Note that this only makes sense because
        # the entire model is fine-tuned.
        tokens += [sep_token]
        label_ids += [pad_token_label_id]
        if sep_token_extra:
            # roberta uses an extra separator b/w pairs of sentences
            tokens += [sep_token]
            label_ids += [pad_token_label_id]
        segment_ids = [sequence_a_segment_id] * len(tokens)

        if cls_token_at_end:
            tokens += [cls_token]
            label_ids += [pad_token_label_id]
            segment_ids += [cls_token_segment_id]
        else:
            tokens = [cls_token] + tokens
            label_ids = [pad_token_label_id] + label_ids
            segment_ids = [cls_token_segment_id] + segment_ids

        input_ids = tokenizer.convert_tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)

        # Zero-pad up to the sequence length.
        padding_length = max_seq_length - len(input_ids)
        if pad_on_left:
            input_ids = ([pad_token] * padding_length) + input_ids
            input_mask = ([0 if mask_padding_with_zero else 1] * padding_length) + input_mask
            segment_ids = ([pad_token_segment_id] * padding_length) + segment_ids
            label_ids = ([pad_token_label_id] * padding_length) + label_ids
        else:
            input_ids += [pad_token] * padding_length
            input_mask += [0 if mask_padding_with_zero else 1] * padding_length
            segment_ids += [pad_token_segment_id] * padding_length
            label_ids += [pad_token_label_id] * padding_length

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length
        assert len(label_ids) == max_seq_length

        if ex_index < 5:
            logger.info("*** Example ***")
            logger.info("guid: %s", example.guid)
            logger.info("tokens: %s", " ".join([str(x) for x in tokens]))
            logger.info("input_ids: %s", " ".join([str(x) for x in input_ids]))
            logger.info("input_mask: %s", " ".join([str(x) for x in input_mask]))
            logger.info("segment_ids: %s", " ".join([str(x) for x in segment_ids]))
            logger.info("label_ids: %s", " ".join([str(x) for x in label_ids]))

        if "token_type_ids" not in tokenizer.model_input_names:
            segment_ids = None
        if is_distil_bert:
            features.append(
                InputFeatures(
                    input_ids=input_ids,
                    attention_mask=input_mask,
                    label_ids=label_ids,
                    document_id=example.guid,
                )
            )
        else:
            features.append(
                InputFeatures(
                    input_ids=input_ids,
                    attention_mask=input_mask,
                    token_type_ids=segment_ids,
                    label_ids=label_ids,
                    document_id=example.guid,
                )
            )
    return features


def get_labels(path: str) -> List[str]:
    if path:
        with open(path, "r") as f:
            labels = f.read().splitlines()
        if "O" not in labels:
            labels = ["O"] + labels
        return labels
    else:
        return [
            "O",
            "B-MISC",
            "I-MISC",
            "B-PER",
            "I-PER",
            "B-ORG",
            "I-ORG",
            "B-LOC",
            "I-LOC",
        ]



#!/usr/bin/env python3

import argparse
import typing
import unicodedata
from pathlib import Path
from xml.etree import ElementTree as ET

import tqdm

import bunkai.constant


def get_sentence(node) -> typing.Iterator[str]:
    if node.tag == "sentence":
        yield bunkai.constant.METACHAR_SENTENCE_BOUNDARY
    for luw in node:
        if luw.tag == "webBr":
            yield bunkai.constant.METACHAR_LINE_BREAK
            continue
        elif luw.tag.lower() != "luw":
            yield from get_sentence(luw)
            continue

        for fragment in luw:
            if fragment.tag in ["webBr", "br"]:
                yield bunkai.constant.METACHAR_LINE_BREAK
                continue
            elif fragment.tag.startswith("note"):
                continue
            elif fragment.tag in ["sampling", "fraction", "info"]:
                continue
            elif fragment.tag.lower() != "suw":
                raise NotImplementedError(fragment.tag)
            elif fragment.text is None:
                for unit in fragment:
                    if unit.tag == "sampling":
                        pass
                    elif unit.tag == "correction":
                        yield unit.get("originalText")
                    elif unit.tag not in [
                        "ruby",
                        "quote",
                        "enclosedCharacter",
                        "delete",
                        "subScript",
                        "superScript",
                    ]:
                        raise NotImplementedError(unit.tag)
                    else:
                        yield unit.text
                continue
            yield fragment.text


def operate_article(parent) -> typing.Iterator[str]:
    for node in parent:
        if node.tag in [
            "rejectedBlock",
            "info",
            "titleBlock",
            "abstract",
            "authorsData",
            "figureBlock",
        ]:
            continue
        elif node.tag in ["webBr", "br"]:
            yield bunkai.constant.METACHAR_LINE_BREAK
        elif node.tag == "sentence":
            yield from get_sentence(node)
        else:
            yield from operate_article(node)


def operation(data: str) -> typing.Iterator[str]:
    root = ET.fromstring(data)
    for child in root:
        if child.tag != "article":
            continue
        for i, t in enumerate(operate_article(child)):
            if i == 0:
                yield t.lstrip(bunkai.constant.METACHAR_SENTENCE_BOUNDARY)
            else:
                yield t
        yield "\n"


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument("--input", "-i", type=Path, required=True)
    oparser.add_argument("--output", "-o", type=Path, required=True)
    oparser.add_argument("--nonfkc", action="store_true")
    return oparser.parse_args()


def main() -> None:
    opts = get_opts()
    if opts.input.is_dir():
        opts.output.mkdir(exist_ok=True, parents=True)
        targets = [n for n in opts.input.iterdir()]
        for fname in tqdm.tqdm(targets, leave=False):
            with fname.open() as inf, opts.output.joinpath(f"{fname.stem}.txt").open("w") as outf:
                for string in operation(inf.read()):
                    if not opts.nonfkc:
                        string = unicodedata.normalize("NFKC", string)
                    outf.write(string)
    else:
        with opts.input.open() as inf, opts.output.open("w") as outf:
            for string in operation(inf.read()):
                if not opts.nonfkc:
                    string = unicodedata.normalize("NFKC", string)
                outf.write(string)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from typing import List

from janome.tokenizer import Tokenizer

from bunkai.base.annotation import Annotations, SpanAnnotation, TokenResult
from bunkai.base.annotator import Annotator


class MorphAnnotatorJanome(Annotator):
    def __init__(self):
        super().__init__(rule_name=self.__class__.__name__)
        self.tokenizer = Tokenizer()

    def __generate(self, text: str) -> List[SpanAnnotation]:
        tokenizer_result = self.tokenizer.tokenize(text)
        span_ann = []
        __start_index = 0
        for t_obj in tokenizer_result:
            __pos = t_obj.part_of_speech.split(",")
            __length = len(t_obj.surface)
            token = TokenResult(
                node_obj=t_obj,
                tuple_pos=__pos,
                word_stem=t_obj.base_form,
                word_surface=t_obj.surface,
            )
            span_ann.append(
                SpanAnnotation(
                    rule_name=self.rule_name,
                    start_index=__start_index,
                    end_index=__start_index + __length,
                    split_string_type="janome",
                    split_string_value="token",
                    args={"token": token},
                )
            )
            __start_index += __length
        else:
            if __start_index < len(text) and text[__start_index : len(text)] == "\n":
                # 末尾が改行のケースで改行記号を手動で追加する。
                token = TokenResult(
                    node_obj=None,
                    tuple_pos=("記号", "空白", "*", "*"),
                    word_stem="\n",
                    word_surface="\n",
                )
                span_ann.append(
                    SpanAnnotation(
                        rule_name=self.rule_name,
                        start_index=__start_index,
                        end_index=len(text),
                        split_string_type="janome",
                        split_string_value="token",
                        args={"token": token},
                    )
                )

            return span_ann

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        anns = self.__generate(original_text)
        spans.add_annotation_layer(self.rule_name, anns + list(spans.flatten()))
        return spans



#!/usr/bin/env python3
from typing import Dict, List, Type

from bunkai.algorithm.tsunoda_sbd.annotator.basic_annotator import BasicRule
from bunkai.algorithm.tsunoda_sbd.annotator.morph_annotator_janome import MorphAnnotatorJanome
from bunkai.base.annotation import Annotations, SpanAnnotation, TokenResult
from bunkai.base.annotator import Annotator


class ExceptionParticle(Annotator):
    def __init__(self, morph_annotator_class: Type[MorphAnnotatorJanome]):
        super().__init__(rule_name=self.__class__.__name__)
        self.morph_annotator_class = morph_annotator_class

    @staticmethod
    def is_exception_particle(
        original_text: str,
        start_index: int,
        end_index: int,
        index2token_obj: Dict[int, TokenResult],
    ) -> bool:
        """
        形態素解析の結果、基本分割文字列の後ろが助詞だった場合は 分割を行わない.

        例: 合宿免許? の若者さん達でしょうか / スタッフ? と話し込み .
        """
        __next_end_index = end_index
        if __next_end_index not in index2token_obj:
            return False
        else:
            token_obj = index2token_obj[__next_end_index]
            if token_obj.tuple_pos[0] == "助詞":
                return True
            else:
                return False

    def __generate(self, anns: List[SpanAnnotation]) -> Dict[int, TokenResult]:
        index2tokens = {}
        __start_index = 0
        __tokenizer_anns = [ann for ann in anns if ann.rule_name == self.morph_annotator_class.__name__]
        __processed = []
        for ann in __tokenizer_anns:
            t_obj = ann.args["token"]  # type: ignore
            if t_obj in __processed:
                continue
            __length = len(t_obj.word_surface)
            for __i in range(__start_index, __start_index + __length):
                index2tokens[__i] = t_obj
            __start_index += __length
            __processed.append(t_obj)
        else:
            return index2tokens

    def annotate(
        self,
        original_text: str,
        spans: Annotations,
    ) -> Annotations:
        index2token_obj = self.__generate(list(spans.get_annotation_layer(self.morph_annotator_class.__name__)))

        __return_span_ann = []
        for __s in spans.name2spans[BasicRule.__name__]:
            if self.is_exception_particle(
                original_text,
                __s.start_index,
                __s.end_index,
                index2token_obj=index2token_obj,
            ):
                continue
            else:
                __s.rule_name = self.rule_name
                __return_span_ann.append(__s)

        spans.add_annotation_layer(self.rule_name, __return_span_ann)
        return spans



#!/usr/bin/env python3

import argparse
import sys
import tempfile
import typing
import zipfile
from pathlib import Path

import bunkai.constant
from bunkai import __version__
from bunkai.algorithm.bunkai_sbd.bunkai_sbd import BunkaiSentenceBoundaryDisambiguation
from bunkai.algorithm.tsunoda_sbd.tsunoda_sbd import TsunodaSentenceBoundaryDisambiguation

DEFAULT_ALGORITHM = "bunkai"
algorithm2class: typing.Dict[str, typing.Type] = {
    DEFAULT_ALGORITHM: BunkaiSentenceBoundaryDisambiguation,
    "tsunoda": TsunodaSentenceBoundaryDisambiguation,
}


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument(
        "--algorithm",
        "-a",
        default=DEFAULT_ALGORITHM,
        choices=sorted(list(algorithm2class.keys())),
    )
    oparser.add_argument(
        "--input",
        "-i",
        type=Path,
        default="/dev/stdin",
        required=False,
    )
    oparser.add_argument(
        "--output",
        "-o",
        type=Path,
        default="/dev/stdout",
        required=False,
    )
    oparser.add_argument(
        "--model",
        "-m",
        type=Path,
    )
    oparser.add_argument(
        "--setup",
        action="store_true",
        help="Setup a model file",
    )
    oparser.add_argument(
        "--ma",
        action="store_true",
        help="Print Morphological analyses result",
    )
    oparser.add_argument(
        "--version",
        "-v",
        action="store_true",
        help="Print version",
    )
    return oparser.parse_args()


def is_install_with_lb() -> bool:
    try:
        import numpy  # noqa: F401
        import torch  # noqa: F401
        import transformers  # noqa: F401
    except ImportError:
        return False
    else:
        return True


def run(
    annotator,
    _text: str,
    ma: bool = False,
) -> typing.Iterator[str]:
    assert "\n" not in _text
    assert bunkai.constant.METACHAR_SENTENCE_BOUNDARY not in _text

    text: str = _text.replace(
        bunkai.constant.METACHAR_LINE_BREAK,
        "\n",
    )

    if ma:
        annotation_obj = annotator.eos(text)
        tokens = annotation_obj.get_morph_analysis()
        end_indices = set([s_a.end_index for s_a in annotation_obj.get_final_layer()])

        position: int = 0
        for (
            idx,
            token,
        ) in enumerate(tokens):
            prev_position: int = position
            if token.node_obj is None or token.word_surface == "\n":
                yield bunkai.constant.METACHAR_LINE_BREAK
                position += 1
            else:
                yield f"{token.word_surface}\t"
                node = token.node_obj
                yield f"{node.part_of_speech},{node.infl_type},{node.infl_form}"
                yield f",{node.base_form},{node.reading},{node.phonetic}"
                position += len(token.word_surface)
            yield "\n"
            for p in range(
                prev_position,
                position,
            ):
                if p + 1 in end_indices:
                    yield "EOS\n"
    else:
        __annotator_result: typing.List[int] = annotator.find_eos(text)

        last: int = 0
        for (
            idx,
            split_point,
        ) in enumerate(__annotator_result):
            yield text[last:split_point].replace(
                "\n",
                bunkai.constant.METACHAR_LINE_BREAK,
            )
            if idx != len(__annotator_result) - 1:
                yield bunkai.constant.METACHAR_SENTENCE_BOUNDARY
            last = split_point
        yield "\n"


def setup(
    path_model: Path,
    path_in: typing.Optional[Path],
):
    # if bunkai installed without [lb] option, import of lbd.dist fails
    from bunkai.algorithm.lbd.dist import restore

    sys.stderr.write("It takes time to setup. Please be patient.\n")
    with tempfile.TemporaryDirectory() as temp_path:
        import requests

        if path_in is None:
            url: str = "https://github.com/megagonlabs/bunkai/releases/download/v1.1.1/bunkai-model-setup-20210426.zip"
            sys.stderr.write(f"Downloading from {url}\n")
            res = requests.get(url)
            assert res.status_code == 200
            path_in = Path(temp_path).joinpath("setup.zip")
            with path_in.open("wb") as fout:
                fout.write(res.content)

        sys.stderr.write("Extracting...\n")
        path_src = Path(temp_path).joinpath("src")
        path_src.mkdir(
            exist_ok=True,
            parents=True,
        )
        with zipfile.ZipFile(path_in) as zipf:
            zipf.extractall(path_src)

        sys.stderr.write("Writing...\n")
        files = [f for f in path_src.iterdir()]
        assert len(files) == 1
        restore(
            files[0],
            path_model,
        )
    sys.stderr.write("Done!\n")


def main() -> None:
    opts = get_opts()

    if opts.version:
        print(f"Bunkai {__version__}")
        return

    install_with_lb = is_install_with_lb()
    if opts.model is not None and not install_with_lb:
        raise ImportError(
            "To use model, you need numpy, transformers, and torch. "
            "It is recommended to install bunkai with `pip install bunkai[lb]`."
        ) from None

    if opts.setup:
        assert opts.model is not None, "--model should be given"
        pin = opts.input
        if opts.input == Path("/dev/stdin"):
            pin = None
        setup(
            opts.model,
            pin,
        )
        return

    cls = algorithm2class[opts.algorithm]
    annotator = cls(path_model=opts.model)
    warned: bool = False

    with opts.input.open() as inf, opts.output.open("w") as outf:
        for line in inf:
            ol: str = line[:-1]
            if bunkai.constant.METACHAR_SENTENCE_BOUNDARY in ol:
                ol = ol.replace(
                    bunkai.constant.METACHAR_SENTENCE_BOUNDARY,
                    "",
                )
                if not warned:
                    sys.stderr.write(
                        "\033[91m"
                        f"[Warning] All {bunkai.constant.METACHAR_SENTENCE_BOUNDARY} will be removed for input\n"
                        "\033[0m"
                    )
                    warned = True

            for op in run(
                annotator,
                ol,
                opts.ma,
            ):
                outf.write(op)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import json
import logging
import shutil
from pathlib import Path
from typing import List, Optional

import numpy as np
import torch
from transformers.models.auto.configuration_auto import AutoConfig
from transformers.models.bert.modeling_bert import BertForTokenClassification
from transformers.utils.hub import cached_file

_NAME_UPDATER_DIR: str = "up"
_DIST_VERSION: str = "1.1.0"
_NAME_VERSION: str = "version.json"


def to_numpy(v):
    return v.to("cpu").detach().numpy().copy()


def to_tensor(v):
    return torch.from_numpy(v.astype(np.float32)).clone()


_COPY_FILES: List[str] = [
    "added_tokens.json",
    "special_tokens_map.json",
    "tokenizer_config.json",
    "bunkai.json",
    "labels.txt",
    "config.json",
]


def store_copy(path_in: Path, path_out: Path):
    for t in _COPY_FILES:
        shutil.copy2(path_in.joinpath(t), path_out.joinpath(t))


def store_updater(path_in: Path, base_model: str, path_out: Path):
    logging.disable(logging.WARN)
    orig_model = BertForTokenClassification.from_pretrained(base_model)
    assert isinstance(orig_model, BertForTokenClassification)
    new_model = BertForTokenClassification.from_pretrained(path_in)
    assert isinstance(new_model, BertForTokenClassification)

    path_out_up = path_out.joinpath(_NAME_UPDATER_DIR)
    path_out_up.mkdir(parents=True, exist_ok=True)
    for target, _orgv in orig_model.named_parameters():
        orgv = to_numpy(_orgv)
        newv = to_numpy(new_model.state_dict()[target])
        if newv.shape != orgv.shape:
            if len(orgv.shape) == 2:
                z = np.zeros((newv.shape[0] - orgv.shape[0], orgv.shape[1]))
            elif len(orgv.shape) == 1:
                z = np.zeros((newv.shape[0] - orgv.shape[0],))
            else:
                raise NotImplementedError
            orgv = np.concatenate((orgv, z), axis=0)
        d = newv - orgv
        np.savez_compressed(path_out_up.joinpath(f"{target}.npz"), d)


def store_version(path_out: Path):
    with path_out.open("w") as outf:
        json.dump({"version": _DIST_VERSION}, outf, sort_keys=True)


def store(path_in: Path, path_out: Path) -> None:
    path_out.mkdir(parents=True, exist_ok=True)

    store_version(path_out.joinpath(_NAME_VERSION))
    store_copy(path_in, path_out)

    with path_in.joinpath("bunkai.json").open() as inf:
        base_model: Optional[str] = json.load(inf).get("base_model")
    if base_model is not None:
        store_updater(path_in, base_model, path_out)


def update(path_in: Path, base_model: str, path_out: Path):
    logging.disable(logging.WARN)
    orig_model = BertForTokenClassification.from_pretrained(base_model)
    assert isinstance(orig_model, BertForTokenClassification)

    osd = orig_model.state_dict()
    for f in path_in.joinpath(_NAME_UPDATER_DIR).iterdir():
        target: str = f.name[: -len(".npz")]
        vals = np.load(f)["arr_0"]
        orgv = to_numpy(osd[target])
        if len(orgv.shape) == 2:
            z = np.zeros((vals.shape[0] - orgv.shape[0], orgv.shape[1]))
        elif len(orgv.shape) == 1:
            z = np.zeros((vals.shape[0] - orgv.shape[0],))
        else:
            raise NotImplementedError

        orgv = np.concatenate((orgv, z), axis=0)
        orgv += vals
        orgv2 = torch.nn.Parameter(to_tensor(orgv))  # type: ignore
        osd[target] = orgv2

    new_model = BertForTokenClassification(config=AutoConfig.from_pretrained(path_in))
    new_model.load_state_dict(osd)  # type: ignore
    new_model.save_pretrained(path_out)  # type: ignore


def check_version(path_in: Path) -> bool:
    with path_in.open() as inf:
        version: str = json.load(inf).get("version", "")
    return version == _DIST_VERSION


def restore(path_in: Path, path_out: Path) -> None:
    if not check_version(path_in.joinpath(_NAME_VERSION)):
        raise KeyError

    path_out.mkdir(parents=True, exist_ok=True)

    with path_in.joinpath("bunkai.json").open() as inf:
        base_model: Optional[str] = json.load(inf).get("base_model")

    shutil.copy2(cached_file(base_model, "vocab.txt"), path_out.joinpath("vocab.txt"))  # type: ignore

    if base_model is not None:
        update(path_in, base_model, path_out)

    # config.json will be overwritten
    store_copy(path_in, path_out)


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument("--input", "-i", type=Path, required=True)
    oparser.add_argument("--output", "-o", type=Path, required=True)
    oparser.add_argument("--restore", action="store_true")
    return oparser.parse_args()


def main() -> None:
    opts = get_opts()
    if opts.restore:
        restore(opts.input, opts.output)
    else:
        store(opts.input, opts.output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import dataclasses
import json
import sys
import typing
from pathlib import Path

from dataclasses_json import DataClassJsonMixin
from torch import nn
from torch.utils.data.dataset import Dataset

from bunkai.algorithm.lbd.corpus import LABEL_NSEP, LABEL_OTHER, LABEL_SEP
from bunkai.base.annotation import Tokens
from bunkai.third.utils_ner import InputFeatures, convert_examples_to_features


def convert(datapath: Path, data_dir: Path, name: str):
    train_data_path = data_dir.joinpath(f"{name}.txt")
    with datapath.open() as inf, train_data_path.open("w") as outf:
        outf.write("-DOCSTART-\n")
        for line in inf:
            tokens = Tokens.from_json(line)
            for idx, span in enumerate(tokens.spans):
                outf.write(f"{span} {tokens.labels[idx]}\n")
            outf.write("\n")


@dataclasses.dataclass
class BunkaiConfig(DataClassJsonMixin):
    max_seq_length: int
    base_model: str


def prepare_config(
    train_path: Path,
    dev_path: Path,
    modelpath: Path,
    max_seq_length: int,
    base_model: str,
    num_train_epochs: int,
) -> Path:
    json_config = {
        "data_dir": "",
        "labels": "",
        "output_dir": "",
        "model_name_or_path": base_model,
        "max_seq_length": max_seq_length,
        "num_train_epochs": num_train_epochs,
        "per_device_train_batch_size": 32,
        "save_steps": 750,
        "seed": 1,
        "do_train": True,
        "do_eval": False,
        "do_predict": False,
        "overwrite_output_dir": True,
        "overwrite_cache": True,
    }

    modelpath.mkdir(exist_ok=True, parents=True)
    path_config_json = modelpath.joinpath("config.json")

    data_dir = modelpath.joinpath("data")
    data_dir.mkdir(exist_ok=True, parents=True)

    out_dir = modelpath.joinpath("out")
    out_dir.mkdir(exist_ok=True, parents=True)

    label_path = out_dir.joinpath("labels.txt")
    with out_dir.joinpath("bunkai.json").open("w") as f:
        json.dump(
            BunkaiConfig(
                base_model=base_model,
                max_seq_length=max_seq_length,
            ).to_dict(),
            f,
            indent=4,
            sort_keys=True,
        )

    json_config["data_dir"] = str(data_dir.absolute())
    json_config["labels"] = str(label_path.absolute())
    json_config["output_dir"] = str(out_dir.absolute())

    with path_config_json.open("w") as f:
        json.dump(json_config, f, indent=4, sort_keys=True)
    with label_path.open("w") as f:
        f.write(f"{LABEL_OTHER}\n{LABEL_SEP}\n{LABEL_NSEP}\n")

    convert(train_path, data_dir, "train")
    if dev_path is not None:
        convert(dev_path, data_dir, "dev")
    return path_config_json


def train(
    train_path: Path,
    dev_path: Path,
    modelpath: Path,
    max_seq_length: int,
    base_model: str,
    num_train_epochs: int,
):
    path_config_json = prepare_config(train_path, dev_path, modelpath, max_seq_length, base_model, num_train_epochs)
    sys.argv = ["", str(path_config_json)]
    import bunkai.third.run_ner

    bunkai.third.run_ner.main()


class MyDataset(Dataset):
    features: typing.List[InputFeatures]
    pad_token_label_id: int = nn.CrossEntropyLoss().ignore_index

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i) -> InputFeatures:
        return self.features[i]

    def __init__(self, examples, labels, max_seq_length: int, tokenizer, is_xlnet: bool):
        self.features = convert_examples_to_features(
            examples,
            labels,
            max_seq_length,
            tokenizer,
            cls_token_at_end=is_xlnet,
            # xlnet has a cls token at the end
            cls_token=tokenizer.cls_token,
            cls_token_segment_id=2 if is_xlnet else 0,
            sep_token=tokenizer.sep_token,
            pad_on_left=bool(tokenizer.padding_side == "left"),
            pad_token=tokenizer.pad_token_id,
            pad_token_segment_id=tokenizer.pad_token_type_id,
            pad_token_label_id=self.pad_token_label_id,
        )


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument("--input", "-i", type=Path, required=True)
    oparser.add_argument("--dev", type=Path)
    oparser.add_argument("--model", "-m", type=Path, required=True)
    oparser.add_argument("--seq", default=320, type=int)
    oparser.add_argument("--epoch", default=30, type=int)
    oparser.add_argument(
        "--base",
        default="bandainamco-mirai/distilbert-base-japanese",
        choices=[
            "cl-tohoku/bert-base-japanese-whole-word-masking",
            "bandainamco-mirai/distilbert-base-japanese",
        ],
    )
    return oparser.parse_args()


def main() -> None:
    opts = get_opts()

    train(
        train_path=opts.input,
        dev_path=opts.dev,
        modelpath=opts.model,
        max_seq_length=opts.seq,
        base_model=opts.base,
        num_train_epochs=opts.epoch,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import dataclasses
from typing import Dict, List, Tuple

from bunkai.algorithm.bunkai_sbd.annotator.basic_annotator import BasicRule
from bunkai.algorithm.bunkai_sbd.annotator.constant import LAYER_NAME_FIRST
from bunkai.algorithm.bunkai_sbd.annotator.emoji_annotator import EmojiAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.emotion_expression_annotator import EmotionExpressionAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.facemark_detector import FaceMarkDetector
from bunkai.algorithm.bunkai_sbd.annotator.linebreak_annotator_compat import LinebreakAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.morph_annotator import MorphAnnotatorJanome
from bunkai.base.annotation import Annotations, SpanAnnotation, TokenResult
from bunkai.base.annotator import AnnotationFilter

DEFAULT_RULE_TARGET = (
    LAYER_NAME_FIRST,
    BasicRule.__name__,
    LinebreakAnnotator.__name__,
    EmojiAnnotator.__name__,
    EmotionExpressionAnnotator.__name__,
    FaceMarkDetector.__name__,
)


@dataclasses.dataclass
class RuleObject(object):
    rule_word_surface: List[str]

    def match(self, current_target_index: int, index2token_obj: Dict[int, TokenResult]):
        for rule_surf in self.rule_word_surface:
            t = index2token_obj.get(current_target_index)
            if t is None:
                return False

            if t.word_surface != rule_surf:
                return False
            current_target_index += len(rule_surf)
        return True


MORPHEMES_AFTER_CANDIDATE: List[RuleObject] = [
    RuleObject(rule_word_surface=["て"]),
    RuleObject(rule_word_surface=["の"]),
    RuleObject(rule_word_surface=["と"]),
    RuleObject(rule_word_surface=["って"]),
    RuleObject(rule_word_surface=["という"]),
    RuleObject(rule_word_surface=["に"]),
    RuleObject(rule_word_surface=["など"]),
    RuleObject(rule_word_surface=["くらい", "の"]),
    RuleObject(rule_word_surface=["くらい", "です"]),
    RuleObject(rule_word_surface=["くらい", "でし"]),
    RuleObject(rule_word_surface=["も", "あり"]),
    RuleObject(rule_word_surface=["ほど", "でし"]),
]


class IndirectQuoteExceptionAnnotator(AnnotationFilter):
    def __init__(self, rule_targets: Tuple[str, ...] = DEFAULT_RULE_TARGET):
        super().__init__(rule_name=self.__class__.__name__)
        self.rule_targets = rule_targets

    @staticmethod
    def is_exception_particle(
        original_text: str,
        start_index: int,
        end_index: int,
        index2token_obj: Dict[int, TokenResult],
    ) -> bool:
        """
        形態素解析の結果、基本分割文字列の後ろが助詞だった場合は 分割を行わない.

        ただし、助詞のすべてのケースで分割するわけではない。ルールを参照して、分割可否を決定する.

        例: 合宿免許? の若者さん達でしょうか / スタッフ? と話し込み.

        :return: True: not SB False: SB.
        """
        __next_end_index = end_index
        # __next_end_indexが最後の文字の場合
        if __next_end_index == len(original_text):
            return False
        # 特殊なルール end_indexの次の文字が改行記号である場合: 改行記号の後の形態素を判定基準にする。
        if original_text[__next_end_index] == "\n":
            while original_text[__next_end_index] == "\n" and __next_end_index + 1 < len(original_text):
                __next_end_index += 1

        if __next_end_index not in index2token_obj:
            return False
        elif any(
            [
                rule_object.match(
                    current_target_index=__next_end_index,
                    index2token_obj=index2token_obj,
                )
                for rule_object in MORPHEMES_AFTER_CANDIDATE
            ]
        ):
            return True
        return False

    def __generate(self, anns: List[SpanAnnotation]) -> Dict[int, TokenResult]:
        index2tokens = {}
        __start_index = 0
        __tokenizer_anns = [ann for ann in anns if ann.rule_name == MorphAnnotatorJanome.__name__]
        __processed = []
        for ann in __tokenizer_anns:
            t_obj = ann.args["token"]  # type: ignore
            if t_obj in __processed:
                continue
            __length = len(t_obj.word_surface)
            for __i in range(__start_index, __start_index + __length):
                index2tokens[__i] = t_obj
            __start_index += __length
            __processed.append(t_obj)
        else:
            return index2tokens

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        index2token_obj = self.__generate(list(spans.get_annotation_layer(MorphAnnotatorJanome.__name__)))

        __return_span_ann = []
        for target_rule_name in self.rule_targets:
            if target_rule_name == LinebreakAnnotator.__name__ and target_rule_name not in spans.name2order:
                continue
            for __s in spans.get_annotation_layer(target_rule_name):
                if self.is_exception_particle(
                    original_text,
                    __s.start_index,
                    __s.end_index,
                    index2token_obj=index2token_obj,
                ):
                    continue
                else:
                    __return_span_ann.append(__s)

        spans.add_annotation_layer(self.rule_name, self.unify_span_annotations(__return_span_ann))
        return spans



#!/usr/bin/env python3
import re

from bunkai.algorithm.tsunoda_sbd.annotator import constant
from bunkai.base.annotation import Annotations, SpanAnnotation
from bunkai.base.annotator import Annotator

RE_SENT_SPLIT = re.compile("[" + constant.PUNCTUATIONS + "]+|" + constant.SYMBOLS)


class BasicRule(Annotator):
    def __init__(self):
        super().__init__(rule_name=self.__class__.__name__)

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        reg_points = RE_SENT_SPLIT.finditer(original_text)

        __return = [
            SpanAnnotation(
                rule_name=self.rule_name,
                start_index=r_obj.regs[0][0],
                end_index=r_obj.regs[0][1],
                split_string_type="regular",
                split_string_value=original_text[r_obj.regs[0][0] : r_obj.regs[0][1]],
            )
            for r_obj in reg_points
        ]
        spans.add_annotation_layer(self.rule_name, __return)
        return spans



#!/usr/bin/env python3
import dataclasses
import itertools
from typing import Any, Dict, Iterator, List, Optional, Tuple

import spans
from dataclasses_json import DataClassJsonMixin


class TokenResult:
    def __init__(
        self,
        node_obj: Any,
        tuple_pos: Tuple[str, ...],
        word_stem: str,
        word_surface: str,
        is_feature=True,
        is_surface=False,
        misc_info=None,
    ):
        self.node_obj = node_obj
        self.word_stem = word_stem
        self.word_surface = word_surface
        self.is_surface = is_surface
        self.is_feature = is_feature
        self.misc_info = misc_info
        self.tuple_pos = tuple_pos

    def __str__(self):
        return self.word_surface


@dataclasses.dataclass
class SpanAnnotation:
    rule_name: Optional[str]
    start_index: int
    end_index: int
    split_string_type: Optional[str]
    split_string_value: Optional[str]
    args: Optional[Dict[str, Any]] = None

    def __str__(self):
        return f"{self.start_index}-{self.end_index}/{self.rule_name}/{self.split_string_value}"

    def __int__(self) -> int:
        return self.end_index

    def __spans(self) -> spans.intrange:
        return spans.intrange(self.start_index, self.end_index)

    def get_spans(self) -> spans.intrange:
        return self.__spans()

    def update_rule_name(self, new_rule_name: str) -> None:
        self.rule_name = new_rule_name


@dataclasses.dataclass
class Annotations:
    annotator_forward: Optional[str] = None
    name2spans: Dict[str, List[SpanAnnotation]] = dataclasses.field(default_factory=dict)
    name2order: Dict[str, int] = dataclasses.field(default_factory=dict)
    current_order: int = 0

    def add_annotation_layer(self, annotator_name: str, annotations: List[SpanAnnotation]) -> None:
        self.name2spans[annotator_name] = annotations
        self.name2order[annotator_name] = self.current_order
        self.annotator_forward = annotator_name
        self.current_order += 1

    def add_flatten_annotations(self, annotations: List[SpanAnnotation]):
        self.name2spans = {
            str(r): list(g_obj)
            for r, g_obj in itertools.groupby(
                sorted(annotations, key=lambda a: a.rule_name),  # type: ignore
                key=lambda a: a.rule_name,
            )
        }

    def flatten(self) -> Iterator[SpanAnnotation]:
        return itertools.chain.from_iterable(self.name2spans.values())

    def get_final_layer(self) -> List[SpanAnnotation]:
        if self.annotator_forward is None:
            return []
        else:
            return self.name2spans[self.annotator_forward]

    def get_annotation_layer(self, layer_name: str) -> Iterator[SpanAnnotation]:
        assert layer_name in list(self.name2spans.keys()), f"{layer_name} not in analysis layers."
        span_anns = {str(ann): ann for ann in itertools.chain.from_iterable(self.name2spans.values())}
        for ann in span_anns.values():
            if ann.rule_name is not None and ann.rule_name == layer_name:
                yield ann
        return

    def get_morph_analysis(self, name_annotation_layer: str = "MorphAnnotatorJanome") -> Iterator[TokenResult]:
        """Get Tokens analysis from Janome."""
        assert name_annotation_layer in self.name2spans, f"{name_annotation_layer} not in annotation layer."
        for span_ann in self.name2spans[name_annotation_layer]:
            if span_ann.rule_name == name_annotation_layer:
                ret = span_ann.args["token"]  # type: ignore
                assert isinstance(ret, TokenResult)
                yield ret
        return

    def available_layers(self) -> List[str]:
        return list(self.name2spans.keys())


@dataclasses.dataclass
class Tokens(DataClassJsonMixin):
    meta: Dict[str, Any] = dataclasses.field(default_factory=dict)
    spans: List[str] = dataclasses.field(default_factory=list)
    labels: List[str] = dataclasses.field(default_factory=list)

    def pretty(self, label2surface: Dict[str, str]) -> str:
        rets: List[str] = []
        for idx, span in enumerate(self.spans):
            rets.append(span)
            label = self.labels[idx]
            olabel = label2surface.get(label)
            if olabel:
                rets.append(olabel)
        return "".join(rets)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bunkai',
 'bunkai.algorithm',
 'bunkai.algorithm.bunkai_sbd',
 'bunkai.algorithm.bunkai_sbd.annotator',
 'bunkai.algorithm.lbd',
 'bunkai.algorithm.tsunoda_sbd',
 'bunkai.algorithm.tsunoda_sbd.annotator',
 'bunkai.base',
 'bunkai.experiment',
 'bunkai.experiment.convert',
 'bunkai.third']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses-json>=0.5.2',
 'emoji>=2.0.0',
 'emojis>=0.6.0',
 'janome>=0.4.1',
 'more_itertools>=8.6.0',
 'spans>=1.1.0',
 'toml>=0.10.2',
 'tqdm']

extras_require = \
{'lb': ['numpy>=1.16.0',
        'torch>=1.3.0',
        'transformers>=4.22.0',
        'requests>=2.27.1,<3.0.0'],
 'train': ['seqeval>=1.2.2']}

entry_points = \
{'console_scripts': ['bunkai = bunkai.cli:main']}

setup_kwargs = {
    'name': 'bunkai',
    'version': '1.5.7',
    'description': 'Sentence boundary disambiguation tool for Japanese texts',
    'long_description': '# Bunkai\n\n[![PyPI version](https://badge.fury.io/py/bunkai.svg)](https://badge.fury.io/py/bunkai)\n[![Python Versions](https://img.shields.io/pypi/pyversions/bunkai.svg)](https://pypi.org/project/bunkai/)\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)\n[![Downloads](https://pepy.tech/badge/bunkai/week)](https://pepy.tech/project/bunkai)\n\n[![CI](https://github.com/megagonlabs/bunkai/actions/workflows/ci.yml/badge.svg)](https://github.com/megagonlabs/bunkai/actions/workflows/ci.yml)\n[![Typos](https://github.com/megagonlabs/bunkai/actions/workflows/typos.yml/badge.svg)](https://github.com/megagonlabs/bunkai/actions/workflows/typos.yml)\n[![CodeQL](https://github.com/megagonlabs/bunkai/actions/workflows/codeql-analysis.yml/badge.svg)](https://github.com/megagonlabs/bunkai/actions/workflows/codeql-analysis.yml)\n[![Maintainability](https://api.codeclimate.com/v1/badges/640b02fa0164c131da10/maintainability)](https://codeclimate.com/github/megagonlabs/bunkai/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/640b02fa0164c131da10/test_coverage)](https://codeclimate.com/github/megagonlabs/bunkai/test_coverage)\n[![markdownlint](https://img.shields.io/badge/markdown-lint-lightgrey)](https://github.com/markdownlint/markdownlint)\n[![jsonlint](https://img.shields.io/badge/json-lint-lightgrey)](https://github.com/dmeranda/demjson)\n[![yamllint](https://img.shields.io/badge/yaml-lint-lightgrey)](https://github.com/adrienverge/yamllint)\n\nBunkai is a sentence boundary (SB) disambiguation tool for Japanese texts.  \n    Bunkaiは日本語文境界判定器です．\n\n## Quick Start\n\n### Install\n\n```console\n$ pip install -U bunkai\n```\n\n### Disambiguation without Models\n\n```console\n$ echo -e \'宿を予約しました♪!まだ2ヶ月も先だけど。早すぎかな(笑)楽しみです★\\n2文書目の先頭行です。▁改行はU+2581で表現します。\' \\\n    | bunkai\n宿を予約しました♪!│まだ2ヶ月も先だけど。│早すぎかな(笑)│楽しみです★\n2文書目の先頭行です。▁│改行はU+2581で表現します。\n```\n\n- Feed a document as one line by using ``▁`` (U+2581) for line breaks.  \n    1行は1つの文書を表します．文書中の改行は ``▁`` (U+2581) で与えてください．\n- The output shows sentence boundaries with ``│`` (U+2502).  \n    出力では文境界は``│`` (U+2502) で表示されます．\n\n### Disambiguation for Line Breaks with a Model\n\nIf you want to disambiguate sentence boundaries for line breaks, please add a ``--model`` option with the path to the model.  \n    改行記号に対しても文境界判定を行いたい場合は，``--model``オプションを与える必要があります．\n\nFirst, please install extras to use ``--model`` option.  \n    ``--model``オプションを利用するために、まずextraパッケージをインストールしてください．\n\n```console\n$ pip install -U \'bunkai[lb]\'\n```\n\nSecond, please setup a model. It will take some time.  \n    次にモデルをセットアップする必要があります．セットアップには少々時間がかかります．\n\n```console\n$ bunkai --model bunkai-model-directory --setup\n```\n\nThen, please designate the directory.  \n    そしてモデルを指定して動かしてください．\n\n```console\n$ echo -e "文の途中で改行を▁入れる文章ってありますよね▁それも対象です。" | bunkai --model bunkai-model-directory\n文の途中で改行を▁入れる文章ってありますよね▁│それも対象です。\n```\n\n### Morphological Analysis Result\n\nYou can get morphological analysis results with ``--ma`` option.  \n``--ma``オプションを付与すると形態素解析結果が得られます．\n\n```console\n$ echo -e \'形態素解析し▁ます。結果を 表示します！\' | bunkai --ma\n形態素\t名詞,一般,*,*,*,*,形態素,ケイタイソ,ケイタイソ\n解析\t名詞,サ変接続,*,*,*,*,解析,カイセキ,カイセキ\nし\t動詞,自立,*,*,サ変・スル,連用形,する,シ,シ\n▁\nEOS\nます\t助動詞,*,*,*,特殊・マス,基本形,ます,マス,マス\n。\t記号,句点,*,*,*,*,。,。,。\nEOS\n結果\t名詞,副詞可能,*,*,*,*,結果,ケッカ,ケッカ\nを\t助詞,格助詞,一般,*,*,*,を,ヲ,ヲ\n \t記号,空白,*,*,*,*, ,*,*\n表示\t名詞,サ変接続,*,*,*,*,表示,ヒョウジ,ヒョージ\nし\t動詞,自立,*,*,サ変・スル,連用形,する,シ,シ\nます\t助動詞,*,*,*,特殊・マス,基本形,ます,マス,マス\n！\t記号,一般,*,*,*,*,！,！,！\nEOS\n```\n\n### Python Library\n\nYou can also use Bunkai as Python library.  \n  BunkaiはPythonライブラリとしても使えます．\n\n```python\nfrom bunkai import Bunkai\nbunkai = Bunkai()\nfor sentence in bunkai("はい。このようにpythonライブラリとしても使えます！"):\n    print(sentence)\n```\n\nFor more information, see [examples](example).  \n    ほかの例は[examples](example)をご覧ください．\n\n## Documents\n\n- [Documents](docs)\n\n## References\n\n- Yuta Hayashibe and Kensuke Mitsuzawa.\n    Sentence Boundary Detection on Line Breaks in Japanese.\n    Proceedings of The 6th Workshop on Noisy User-generated Text (W-NUT 2020), pp.71-75.\n    November 2020.\n    [[PDF]](https://www.aclweb.org/anthology/2020.wnut-1.10.pdf)\n    [[bib]](https://www.aclweb.org/anthology/2020.wnut-1.10.bib)\n\n## License\n\nApache License 2.0\n',
    'author': 'Yuta Hayashibe',
    'author_email': 'hayashibe@megagon.ai',
    'maintainer': 'Yuta Hayashibe',
    'maintainer_email': 'hayashibe@megagon.ai',
    'url': 'https://github.com/megagonlabs/bunkai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
import typing
from typing import Iterator, List

from bunkai.algorithm.tsunoda_sbd.annotator import (
    BasicRule,
    ExceptionNo,
    ExceptionNumeric,
    ExceptionParentheses,
    ExceptionParticle,
    MorphAnnotatorJanome,
)
from bunkai.base.annotation import Annotations, SpanAnnotation
from bunkai.base.annotator import AnnotatorPipeline, RuleOrderException, SentenceBoundaryDisambiguator


class TsunodaPipeline(AnnotatorPipeline):
    def check(self) -> bool:
        # MorphAnnotator must be before ExceptionParticle
        order_morph_annotator = 0
        order_exception_particle = 0
        for __i, a in enumerate(self.pipeline):
            if isinstance(a, MorphAnnotatorJanome):
                order_morph_annotator = __i
            elif isinstance(a, ExceptionParticle):
                order_exception_particle = __i

        if order_morph_annotator > order_exception_particle:
            raise RuleOrderException(
                f"MorphAnnotator at {order_morph_annotator} must be"
                f" before ExceptionParticle at {order_exception_particle}"
            )

        return True


class TsunodaSentenceBoundaryDisambiguation(SentenceBoundaryDisambiguator):
    def __init__(self, *, path_model: typing.Any = None):
        morph_annotator = MorphAnnotatorJanome()  # type: ignore
        particle_annotator = ExceptionParticle(MorphAnnotatorJanome)

        self.pipeline = TsunodaPipeline(
            [
                BasicRule(),
                morph_annotator,
                particle_annotator,
                ExceptionNumeric(),
                ExceptionNo(),
                ExceptionParentheses(),
            ]
        )
        super().__init__()

    def eos(self, text: str) -> Annotations:
        annotations = Annotations()
        annotations.add_annotation_layer(
            "first",
            [
                SpanAnnotation(
                    rule_name=None,
                    start_index=0,
                    end_index=len(text),
                    split_string_type=None,
                    split_string_value=None,
                )
            ],
        )
        for rule_obj in self.pipeline:
            rule_obj.annotate(text, annotations)
        return annotations

    def find_eos(self, text: str) -> List[int]:
        annotations = self.eos(text)
        end_index = list(sorted(list(set([s_a.end_index for s_a in annotations.get_final_layer()]))))
        if len(end_index) == 0 or end_index[-1] != len(text):
            end_index.append(len(text))
        return end_index

    def __call__(self, text: str) -> Iterator[str]:
        annotations = self.eos(text)
        end_index = sorted(list(set([s_a.end_index for s_a in annotations.get_final_layer()])))
        __start_index = 0
        __end_index = 0
        for e_i in end_index:
            part_sentences = text[__start_index:e_i]
            __start_index = e_i
            __end_index = e_i
            yield part_sentences

        if __end_index < len(text):
            part_sentences = text[__end_index:]
            yield part_sentences
